import io
import os
import sys

import six

from brewtils.errors import ValidationError


class GridfsResolver(object):
    """Bytes-Resolver for GridFS

    Resolvers are meant to be written for specific storage types.
    In this case, we are just simply using the API to stream bytes
    into a file.

    This class is meant to be used transparently to Plugin developers.
    Resolvers respond to two methods:

    * `upload(value)`
    * `download(bytes_parameter, writer)`

    Attributes:
        client: A `brewtils.EasyClient`
    """

    def __init__(self, client):
        self.client = client

    def download(self, bytes_parameter, writer):
        """Download the given bytes parameter.

        Args:
            bytes_parameter: A specific request's parameter value.
            writer: File-like object that has a `write` method.
        """
        self.client.stream_to_sink(bytes_parameter["id"], writer)

    def upload(self, value):
        """Upload the given value to the server if necessary.

        The value can be one of the following:

        1. A dictionary with a storage_type and filename.
        2. A string pointing to a valid filename.
        3. An open file descriptor.

        If you use a dictionary, and include an "id" the resolver will
        assume you have already uploaded the file, and skip doing it for you.

        Args:
            value: Value to upload.

        Returns:
            A valid dictionary to use as a bytes parameter.
        """
        if sys.version_info[0] == 2:
            file_types = (io.IOBase, file)  # noqa: F821
        else:
            file_types = (io.IOBase,)
        if isinstance(value, dict):
            return self._upload_dict(value)
        elif isinstance(value, six.string_types):
            return self._upload_filename(value)
        elif isinstance(value, file_types):
            return self._upload_file_descriptor(value)
        else:
            raise ValidationError(
                "Do not know how to upload bytes type %s" % type(value)
            )

    def _upload_dict(self, value):
        if "id" in value:
            return value

        if "filename" not in value:
            raise ValidationError(
                "When uploading a bytes object as a dictionary, you must include a 'filename' key."
            )

        return self._upload_filename(value["filename"], value.get("desired_filename"))

    def _upload_filename(self, filename, desired_filename=None):
        if not os.path.isfile(filename):
            raise ValidationError(
                "Cannot upload a bytes object if the file does not exist %s" % filename
            )

        desired_filename = desired_filename or os.path.basename(filename)

        with open(filename, "rb") as file_to_upload:
            return self._upload_file_descriptor(file_to_upload, desired_filename)

    def _upload_file_descriptor(self, fd, filename=None):
        if filename is None and hasattr(fd, "name"):
            filename = os.path.basename(fd.name)

        filename = filename or str(fd)
        return self.client.upload_file(fd, filename)



from brewtils.resolvers.gridfs import GridfsResolver
from brewtils.resolvers.parameter import DownloadResolver, UploadResolver

__all__ = ["DownloadResolver", "GridfsResolver", "UploadResolver"]

_resolver_map = {"gridfs": {"class": GridfsResolver}}


def build_resolver_map(easy_client):
    """Builds all the resolvers"""
    resolvers = {}
    for key, options in _resolver_map.items():
        klass = options["class"]
        resolvers[key] = klass(client=easy_client)
    return resolvers



import logging
import os
import shutil

from brewtils.errors import ValidationError


class ParameterResolver(object):
    """Base class for parameter resolution.

    This class is used under-the-hood for various plugin functions.
    Its purpose is to remove all the various cleanup and house keeping
    steps involved in resolving parameters. An example of an unresolved
    parameter is a dictionary which represents a bytes object. In this
    case, the user wants the open file descriptor, not the random
    dictionary that they don't know how to process. The parameter resolver
    helps handle these scenarios by providing the following API::

        with ParameterResolver(request, ["my_file"], resolvers) as resolved_params:
            file_bytes = resolved_params["my_file"].read()

    This is intended for internal use for the plugin class.
    """

    def __init__(self, request, params_to_resolve, resolvers):
        if request.is_ephemeral and params_to_resolve:
            raise ValueError("Cannot resolve parameters for ephemeral requests.")

        self.parameters = request.parameters
        self.params_to_resolve = params_to_resolve
        self.resolvers = resolvers
        self.logger = logging.getLogger(__name__)

    def __enter__(self):
        try:
            return self.resolve_parameters()
        except Exception:
            self.cleanup()
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def resolve_parameters(self):
        if not self.params_to_resolve:
            return self.parameters

        self.pre_resolve()
        return self._recurse_and_resolve(self.params_to_resolve, self.parameters)

    def simple_resolve(self, value):
        raise NotImplementedError("No implementation for 'simple_resolve'")

    def pre_resolve(self):
        pass

    def cleanup(self):
        pass

    def _recurse_and_resolve(self, params_to_resolve, parameters):
        if not isinstance(parameters, dict):
            raise ValueError("Nested bytes found, but the value was not a dictionary.")

        resolved_parameters = {}
        top_level_keys = [k[0] for k in params_to_resolve]
        for key, value in parameters.items():
            index = self._index_of(top_level_keys, key)
            if index == -1:
                resolved_value = value
            elif len(params_to_resolve[index]) == 1 and isinstance(value, list):
                resolved_value = [self.simple_resolve(v) for v in value]
            elif len(params_to_resolve[index]) == 1:
                resolved_value = self.simple_resolve(value)
            else:
                resolved_value = self._recurse_and_resolve(
                    params_to_resolve[index], value
                )
            resolved_parameters[key] = resolved_value
        return resolved_parameters

    @staticmethod
    def _index_of(arr, val):
        try:
            return arr.index(val)
        except ValueError:
            return -1

    def _get_resolver(self, value):
        storage_type = "gridfs"
        if isinstance(value, dict):
            storage_type = value.get("storage_type", storage_type)

        if storage_type not in self.resolvers:
            raise ValidationError("No resolver found for %s" % storage_type)

        return self.resolvers[storage_type]


class DownloadResolver(ParameterResolver):
    def __init__(self, request, params_to_resolve, resolvers, base_directory):
        super(DownloadResolver, self).__init__(request, params_to_resolve, resolvers)
        self.base_directory = base_directory
        self._working_dir = os.path.join(str(self.base_directory), request.id or "")

    def cleanup(self):
        if os.path.isdir(self._working_dir):
            shutil.rmtree(self._working_dir)

    def simple_resolve(self, value):
        resolver = self._get_resolver(value)
        filename = value["filename"]
        full_path = os.path.join(self._working_dir, filename)
        if os.path.exists(full_path):
            full_path = os.path.join(self._working_dir, value["id"])

        with open(full_path, "wb") as file_to_write:
            resolver.download(value, file_to_write)
        return full_path

    def pre_resolve(self):
        if not os.path.isdir(self._working_dir):
            os.makedirs(self._working_dir)


class UploadResolver(ParameterResolver):
    def simple_resolve(self, value):
        resolver = self._get_resolver(value)
        return resolver.upload(value)
